// Test harness for the Socket class.    
//
// Can be run either as the server or the client depending on the command line options.
// Will echo a sequence of letters back and forth from client to server a specified number of times.
// Client controls the size of the buffer sent with each send() call and also how many total bytes
// are sent for each send/recv turn.  The server is single threaded, so only one client can talk at
// a time.

#include <iostream>
#include "SocketClient.h"
#include "SocketServer.h"
#include "HiResTimer.h"
#include "Utils.h"

// Fill a char* with cyclic sequences A-Z
void fillBuffer(char* buffer, int len)
{
	int i = 0;
	for (i = 0; i < len; i++)
		buffer[i] = (char) (i % 26 + 65);
}

int main(int argc, char** argv)
{		
	string	host		= "";
	int		port		= 0;

	// Default values for options that don't need to be set
	unsigned int	cycles		= 1;
	unsigned int	msgSize		= 1000;
	unsigned int	maxSend		= 1000;
	unsigned int	backlog		= 10;

	// Parse the command line switches
	if (!getSwitch(argc, argv, "port", &port))
	{
		cout << "TestSocket" << endl;
		cout << " -backlog   <server: maximum backlog>" << endl;
		cout << " -port      <client/server: port number>" << endl;
		cout << " -host      <client: hostname>" << endl;
		cout << " -msgsize   <client: message size in bytes>" << endl;
		cout << " -maxsend   <client: max bytes to send at one time>" << endl;
		cout << " -cycles    <client: send/recv cycles>" << endl;
		cout << " -printmsg  [dump the buffer for debugging]" << endl;

		cout << endl;

		return 0;
	}

	bool isClient = getSwitch(argc, argv, "host", &host);		
	bool printMsg = getSwitch(argc, argv, "printmsg");

	// Create the flag parameter to send in to our socket constructors
	int flags = 0;

	getSwitch(argc, argv, "cycles", &cycles);
	getSwitch(argc, argv, "maxsend", &maxSend);
	getSwitch(argc, argv, "msgsize", &msgSize);
	getSwitch(argc, argv, "backlog", &backlog);

	// Create some fake data to send
	char* buffer = new char[msgSize];
	fillBuffer(buffer, msgSize);

	unsigned long long sendBytes = 0;

	////////////////////////////////////////// client code /////////////////////////////////////////
	if (isClient)
	{
		cout << "Waiting to talk to " << host << " on port " << port << endl;

		bool good = false;
		SocketClient client(host.c_str(), port, flags, good);	
		int numSent = 0;
		int numRecv = 0;
			
		if (!good)
		{
			cerr << "Failed to create client connection" << endl;
			delete buffer;
			buffer = NULL;
			return 0;
		}

		PeerInfo info = client.GetPeerInfo();
		cout << "IP = " << info.name << ", port = " << info.port << endl;

		// Start by telling the server the client choices for msgSize, maxSend, and cycles.		
		// NOTE: this is somewhat naught as it relies on both client and server being same Endian-ness.
		unsigned int setupMsg[3];
		setupMsg[0] = msgSize;
		setupMsg[1] = maxSend;
		setupMsg[2] = cycles;

		numSent = client.SendBytes((char*) setupMsg, sizeof(unsigned int) * 3);
		cout << "CLIENT: setup message of " << numSent << " bytes" << endl;

		// Wait to hear a one-byte response from server
		numRecv = client.RecvBytes((char*) setupMsg, 1);
		cout << "CLIENT: got setup response of " << numRecv << " bytes" << endl;

		for (unsigned int i = 0; i < cycles; i++)
		{			
			// Send as many times as it takes to get through the buffer
			unsigned int left = msgSize;
			char* current = buffer;			
			cout << i << " send ";
			unsigned int bytes = 0;
			unsigned int packets = 0;
			while (left > 0)
			{
				// Send at most maxSend bytes or up to the end of the buffer
				numSent = client.SendBytes(buffer, min(maxSend, left));
				if (numSent == -1)
				{
					cerr << "Failed to send bytes!" << endl;
					delete buffer;
					buffer = NULL;
					return 0;
				}

				left -= numSent;
				current += numSent;
				packets++;
				bytes += numSent;
			}

			cout << packets << " packets, " << bytes << " bytes" << endl;

			// Receive as many times as it takes to get to the full message length
			cout << i << " recv ";
			left = msgSize;
			current = buffer;
			packets = 0;
			bytes = 0;
			while (left > 0)
			{
				numRecv = client.RecvBytes(buffer, msgSize);
				if (numRecv == -1)
				{
					cerr << "Failed to receive bytes!" << endl;
					delete buffer;
					buffer = NULL;
					return 0;
				}
				left -= numRecv;
				packets++;
				bytes += numRecv;
				// Move the pointer so the next receive fills in from where we left off
				current += numRecv;
			}
			cout << packets << " packets, " << bytes << " bytes" << endl;

			// Optionally we print out the received message
			if (printMsg)
			{
				char ch = buffer[msgSize - 1];
				buffer[msgSize - 1] = '\0';
				cout << buffer << endl;
				buffer[msgSize - 1] = ch;
			}

		}		
	}	
	////////////////////////////////////////// server code /////////////////////////////////////////
	else 
	{
		bool good = false;
		SocketServer server(port, backlog, flags, good);

		if (!good)
		{	
			cerr << "Failed to create server connection" << endl;
			delete buffer;
			buffer = NULL;
			return 0;		
		}

		while (true)
		{
			Socket* conn = server.Accept();
			if (conn)
			{
				PeerInfo info = server.GetPeerInfo();
				cout << "IP = " << info.name << ", port = " << info.port << endl;

				// Read the initial setup message from client.
				unsigned int setupMsg[3];
				int numRecv		= 0;
				int numSent		= 0;
				unsigned int oldMsgSize	= msgSize;

				numRecv = conn->RecvBytes((char *) setupMsg, sizeof(unsigned int) * 3);
				msgSize = setupMsg[0];
				maxSend = setupMsg[1];
				cycles  = setupMsg[2];

				cout << "SERVER: client wants msgSize = " << msgSize << ", maxSend = " << maxSend << ", cycles = " << cycles << endl;
				
				// Send the client a single byte to let it know we got the setup message
				numSent= conn->SendBytes((char *) setupMsg, 1);
				cout << "SERVER: sent setup response of " << numSent << " bytes" << endl;

				// Resize the buffer if needed
				if (msgSize > oldMsgSize)
				{
					delete buffer;
					buffer = new char[msgSize];
				}

				for (unsigned int i = 0; i < cycles; i++)
				{					
					// Receive as many times as it takes to get to the full message length
					cout << i << " recv ";
					unsigned int left = msgSize;
					char* current = buffer;
					unsigned int packets = 0;
					unsigned int bytes = 0;

					while (left > 0)
					{
						numRecv = conn->RecvBytes(buffer, msgSize);
						if (numRecv == -1)
						{
							cerr << "Failed to receive bytes!" << endl;
							delete buffer;
							buffer = NULL;
							return 0;
						}
						left -= numRecv;
						packets++;
						bytes += numRecv;
						// Move the pointer so the next receive fills in from where we left off
						current += numRecv;
					}
					cout << packets << " packets, " << bytes << " bytes" << endl;
				
					// Send as many times as it takes to get through the buffer
					left = msgSize;
					current = buffer;			
					cout << i << " send ";
					packets = 0;
					bytes = 0;
					while (left > 0)
					{
						// Send at most maxSend bytes or up to the end of the buffer
						numSent = conn->SendBytes(buffer, min(maxSend, left));
						if (numRecv == -1)
						{
							cerr << "Failed to send bytes!" << endl;
							delete buffer;
							buffer = NULL;
							return 0;
						}
						left -= numSent;
						current += numSent;
						packets++;
						bytes += numSent;

					}
					cout << packets << " packets, " << bytes << " bytes" << endl;

				}

				// All done with this client object
				delete conn;
				conn = NULL;
			}
		}
	}

	delete buffer;
	buffer = NULL;

	return 0;
}

